/**
   @author Shin'ichiro Nakaoka
*/

#include "Archive.h"
#include "Item.h"
#include "AppConfig.h"
#include "MessageView.h"
#include <cnoid/ExecutablePath>
#include <cnoid/Referenced>
#include <cnoid/FileUtil>
#include <map>
#include <vector>
#include <boost/signals.hpp>
#include <boost/algorithm/string.hpp>
#include <QRegExp>
#include "gettext.h"

using namespace std;
using namespace boost;
using namespace cnoid;


namespace {
    QRegExp regexVar("^\\$\\{(\\w+)\\}");

    typedef map<ItemPtr, int> ItemToIdMap;
}


namespace cnoid {
    
    class ArchiveSharedData : public Referenced
    {
    public:
        Mapping* directoryVariableMap;
        filesystem::path projectDirPath;
        filesystem::path topDirPath;
        filesystem::path shareDirPath;
        QString topDirString;
        QString shareDirString;

        vector<Item*> idToItems;
        ItemToIdMap itemToIds;

        Item* currentParentItem;

        boost::signal<void()> postProcesses;
    };


    bool findSubDirectoryOfDirectoryVariable
    (ArchiveSharedData* shared, const filesystem::path& path, std::string& out_varName, filesystem::path& out_relativePath)
    {
        Mapping::const_iterator p;
        for(p = shared->directoryVariableMap->begin(); p != shared->directoryVariableMap->end(); ++p){
            Listing* paths = p->second->toListing();
            if(paths){
                for(int i=0; i < paths->size(); ++i){
                    filesystem::path dirPath(paths[i].toString());
                    if(findSubDirectory(dirPath, path, out_relativePath)){
                        out_varName = fromUTF8(p->first);
                        return true;
                    }
                }
            }
        }
        return false;
    }


    void replaceDirectoryVariable(ArchiveSharedData* shared, QString& io_pathString, const QString& varname, int pos, int len)
    {
        Listing* paths = shared->directoryVariableMap->findListing(varname.toStdString());
        if(paths){
            for(int i=0; i < paths->size(); ++i){
                string vpath;
                QString replaced(io_pathString);
                replaced.replace(pos, len, paths[i].toString().c_str());
                filesystem::file_status fstatus = filesystem::status(filesystem::path(replaced.toStdString()));
                if(filesystem::is_directory(fstatus) || filesystem::exists(fstatus)) {
                    io_pathString = replaced;
                    return;
                }
            }
        }
        MessageView::mainInstance()->putln(QString(_("Warning: ${%1} of \"%2\" cannot be expanded !")).arg(varname).arg(io_pathString));
    }
}


ArchivePtr Archive::invalidArchive()
{
    static ArchivePtr invalidArchive_ = new Archive();
    invalidArchive_->type_ = ValueNode::INVALID_NODE;
    return invalidArchive_;
}


Archive::Archive()
{

}


Archive::Archive(int line, int column)
    : Mapping(line, column)
{

}


Archive::~Archive()
{

}


void Archive::initSharedInfo(const std::string& projectFile)
{
    shared = new ArchiveSharedData;

    shared->directoryVariableMap = AppConfig::archive()->openMapping("PathVariables");

    shared->projectDirPath = getAbsolutePath(filesystem::path(projectFile)).parent_path();
    shared->topDirPath = executableTopDirectory();
    shared->shareDirPath = shareDirectory();

    shared->topDirString = executableTopDirectory().c_str();
    shared->shareDirString = shareDirectory().c_str();
    
    shared->currentParentItem = 0;
}


void Archive::inheritSharedInfoFrom(ArchivePtr archive)
{
    shared = archive->shared;
}


// This cannot be signal because the shared instance may not exist
void Archive::addPostProcess(const boost::function<void()>& func) const
{
    if(shared){
        shared->postProcesses.connect(func);
    }
}


void Archive::callPostProcesses()
{
    if(shared){
        shared->postProcesses();
    }
}


Archive* Archive::findSubArchive(const std::string& name)
{
    Mapping* mapping = findMapping(name);
    if(mapping->isValid()){
        Archive* archive = dynamic_cast<Archive*>(mapping);
        if(archive){
            return archive;
        }
    }

    return invalidArchive().get();
}


bool Archive::readRelocatablePath(const std::string& key, std::string& out_value) const
{
    string orgPathString;

    if(read(key, orgPathString)){

        QString pathString(orgPathString.c_str());

        // expand variables in the path
        int pos = regexVar.indexIn(pathString);
        if(pos != -1){
            int len = regexVar.matchedLength();
            if(regexVar.captureCount() > 0){
                QString varname = regexVar.cap(1);
                if(varname == "SHARE"){
                    pathString.replace(pos, len, shared->shareDirString);
                } else if(varname == "PROGRAM_TOP"){
                    pathString.replace(pos, len, shared->topDirString);
                } else {
                    replaceDirectoryVariable(shared.get(), pathString, varname, pos, len);
                }
            }
        }
            
        filesystem::path path(pathString.toStdString());

        if(checkAbsolute(path)){
            out_value = getNativePathString(path);
        } else {
            filesystem::path fullPath = shared->projectDirPath / path;
            if(!path.empty() && (*path.begin() == "..")){
                filesystem::path compact;
                makePathCompact(fullPath, compact);
                out_value = getNativePathString(compact);
            } else {
                out_value = getNativePathString(fullPath);
            }
        } 
        return true;
    }

    return false;
}


/**
   \todo Use integated nested map whose node is a single path element to be more efficient.
*/
bool Archive::writeRelocatablePath(const std::string& key, const std::string& orgPathString)
{
    filesystem::path orgPath(orgPathString);
    filesystem::path relativePath;
    string varName;

    // In the case where the path is originally relative one
    if(!orgPath.is_complete()){
        write(key, getGenericPathString(orgPath), DOUBLE_QUOTED);
        return false;

    } else if(findSubDirectory(shared->projectDirPath, orgPath, relativePath)){
        write(key, getGenericPathString(relativePath), DOUBLE_QUOTED);
        return true;
    
    } else if(findSubDirectoryOfDirectoryVariable(shared.get(), orgPath, varName, relativePath)){
        write(key, string("${") + varName + ("}/") + getGenericPathString(relativePath), DOUBLE_QUOTED);
        return true;

    } else if(findSubDirectory(shared->shareDirPath, orgPath, relativePath)){
        write(key, string("${SHARE}/") + getGenericPathString(relativePath), DOUBLE_QUOTED);
        return true;

    } else if(findSubDirectory(shared->topDirPath, orgPath, relativePath)){
        write(key, string("${PROGRAM_TOP}/") + getGenericPathString(relativePath), DOUBLE_QUOTED);
        return true;

    } else if(findRelativePath(shared->projectDirPath, orgPath, relativePath)){
        write(key, getGenericPathString(relativePath), DOUBLE_QUOTED);
        return true;
    }

    write(key, getGenericPathString(orgPath), DOUBLE_QUOTED);
    return true;
}


void Archive::registerItemId(Item* item, int id)
{
    if(shared){
        if(id >= (signed)shared->idToItems.size()){
            shared->idToItems.resize(id + 1);
        }
        shared->idToItems[id] = item;
        shared->itemToIds[item] = id;
    }
}


/**
   @return -1 if item does not belong to the archive
*/
//int Archive::getItemId(Item* item) const
int Archive::getItemId(ItemPtr item) const
{
    if(shared && item){
        ItemToIdMap::const_iterator p = shared->itemToIds.find(item);
        if(p != shared->itemToIds.end()){
            return p->second;
        }
    }
    return -1;
}


void Archive::writeItemId(const std::string& key, ItemPtr item)
{
    if(item){
        int id = getItemId(item);
        if(id >= 0){
            write(key, id);
        }
    }
}


Item* Archive::findItem(int id) const
{
    if(shared){
        if(id >= 0 && id < (signed)shared->idToItems.size()){
            return shared->idToItems[id];
        }
    }
    return 0;
}


Item* Archive::currentParentItem() const
{
    if(shared){
        return shared->currentParentItem;
    }
    return 0;
}


void Archive::setCurrentParentItem(Item* parentItem)
{
    if(shared){
        shared->currentParentItem = parentItem;
    }
}
