/**
   @author Shin'ichiro Nakaoka
*/

#include "MovieGenerator.h"
#include <cnoid/TimeBar>
#include <cnoid/SceneView>
#include <cnoid/ExtensionManager>
#include <cnoid/MenuManager>
#include <cnoid/Archive>
#include <cnoid/MainWindow>
#include <cnoid/SpinBox>
#include <cnoid/Button>
#include <cnoid/LineEdit>
#include <cnoid/MessageView>
#include <cnoid/Dialog>
#include <cnoid/Separator>
#include <QLabel>
#include <QDialogButtonBox>
#include <QFileDialog>
#include <boost/filesystem.hpp>
#include <boost/bind.hpp>
#include "gettext.h"

using namespace std;
using namespace boost;
using namespace cnoid;


namespace {
    
    class MovieGenerator : public Dialog
    {
    public:
        bool isRecording;
        bool requestStopRecording;
        
        LineEdit directoryEntry;
        PushButton directoryButton;
        LineEdit basenameEntry;
        DoubleSpinBox beginningTimeSpin;
        CheckBox endingTimeCheck;
        DoubleSpinBox endingTimeSpin;
        DoubleSpinBox fpsSpin;
        SpinBox imageWidthSpin;
        SpinBox imageHeightSpin;

        MovieGenerator();
        bool store(Archive& archive);
        void restore(const Archive& archive);
        void showDirectorySelectionDialog();
        void generate();
        void onStopRequest();
        bool doRecordingLoop();
    };
}


void cnoid::initializeMovieGenerator(ExtensionManager* ext)
{
    static MovieGenerator* movieGenerator = 0;

    if(!movieGenerator){
        movieGenerator = ext->manage(new MovieGenerator());

        MenuManager& mm = ext->menuManager();
        mm.setPath("/Tools");
        mm.addItem(N_("Movie Generator"))
            ->sigTriggered().connect(bind(&MovieGenerator::show, movieGenerator));
        
        ext->connectProjectArchiver(
            "MovieGenerator",
            bind(&MovieGenerator::store, movieGenerator, _1),
            bind(&MovieGenerator::restore, movieGenerator, _1));
    }
}


MovieGenerator::MovieGenerator()
{
    setWindowTitle(_("Movie Generator"));
    
    isRecording = false;
    requestStopRecording = false;

    QVBoxLayout* vbox = new QVBoxLayout();
    setLayout(vbox);

    QHBoxLayout* hbox = new QHBoxLayout();
    hbox->addWidget(new QLabel(_("Directory")));
    hbox->addWidget(&directoryEntry);

    QIcon folderIcon = QIcon::fromTheme("folder");
    if(folderIcon.isNull()){
        directoryButton.setText(_("Select"));
    } else {
        directoryButton.setIcon(folderIcon);
    }
    directoryButton.sigClicked().connect(bind(&MovieGenerator::showDirectorySelectionDialog, this));
    hbox->addWidget(&directoryButton);

    hbox->addWidget(new QLabel(_("Basename")));
    basenameEntry.setText("scene");
    hbox->addWidget(&basenameEntry);
    vbox->addStretch();
    vbox->addLayout(hbox);

    hbox = new QHBoxLayout();
    hbox->addWidget(new QLabel(_("Begin")));
    beginningTimeSpin.setDecimals(1);
    beginningTimeSpin.setRange(0.0, 9999.9);
    beginningTimeSpin.setSingleStep(0.1);
    hbox->addWidget(&beginningTimeSpin);

    endingTimeCheck.setText(_("End"));
    hbox->addWidget(&endingTimeCheck);
    
    endingTimeSpin.setDecimals(1);
    endingTimeSpin.setRange(0.0, 9999.9);
    endingTimeSpin.setSingleStep(0.1);
    hbox->addWidget(&endingTimeSpin);

    hbox->addWidget(new QLabel("FPS"));
    fpsSpin.setDecimals(1);
    fpsSpin.setRange(1.0, 9999.9);
    fpsSpin.setValue(30.0);
    fpsSpin.setSingleStep(0.1);
    hbox->addWidget(&fpsSpin);
    hbox->addStretch();
    vbox->addLayout(hbox);
    
    hbox = new QHBoxLayout();
    hbox->addWidget(new QLabel(_("Image size")));
    
    imageWidthSpin.setRange(1, 9999);
    imageWidthSpin.setValue(640);
    hbox->addWidget(&imageWidthSpin);

    hbox->addWidget(new QLabel("x"));
    imageHeightSpin.setRange(1, 9999);
    imageHeightSpin.setValue(480);
    hbox->addWidget(&imageHeightSpin);
    hbox->addStretch();
    vbox->addLayout(hbox);

    vbox->addWidget(new HSeparator);
    QDialogButtonBox* buttonBox = new QDialogButtonBox(this);
    vbox->addWidget(buttonBox);

    PushButton* applyButton = new PushButton(_("&Generate"));
    applyButton->setDefault(true);
    buttonBox->addButton(applyButton, QDialogButtonBox::ActionRole);
    applyButton->sigClicked().connect(bind(&MovieGenerator::generate, this));
    hbox->addWidget(applyButton);

    PushButton* stopButton = new PushButton(_("&Stop"));
    buttonBox->addButton(stopButton, QDialogButtonBox::ActionRole);
    stopButton->sigClicked().connect(bind(&MovieGenerator::onStopRequest, this));
}


bool MovieGenerator::store(Archive& archive)
{
    archive.write("directory", directoryEntry.string());
    archive.write("basename", basenameEntry.string());
    archive.write("begin", beginningTimeSpin.value());
    if(endingTimeCheck.isChecked()){
        archive.write("end", endingTimeSpin.value());
    }
    archive.write("fps", fpsSpin.value());
    archive.write("width", imageWidthSpin.value());
    archive.write("heiht", imageHeightSpin.value());
    return true;
}


void MovieGenerator::restore(const Archive& archive)
{
    directoryEntry.setText(archive.get("directory", directoryEntry.string()));
    basenameEntry.setText(archive.get("basename", basenameEntry.string()));
    beginningTimeSpin.setValue(archive.get("begin", beginningTimeSpin.value()));
    double endingTime;
    if(archive.read("end", endingTime)){
        endingTimeSpin.setValue(endingTime);
        endingTimeCheck.setChecked(true);
    } else {
        endingTimeCheck.setChecked(false);
    }
    fpsSpin.setValue(archive.get("fps", fpsSpin.value()));
    imageWidthSpin.setValue(archive.get("width", imageWidthSpin.value()));
    imageHeightSpin.setValue(archive.get("height", imageHeightSpin.value()));
}


void MovieGenerator::showDirectorySelectionDialog()
{
    QString directory =
        QFileDialog::getExistingDirectory(
            MainWindow::instance(),
            _("Select Directory"),
            directoryEntry.text(),
            0);
    if(!directory.isNull()){
        directoryEntry.setText(directory);
    }
}


void MovieGenerator::generate()
{
    if(doRecordingLoop()){
        hide();
    }
}


void MovieGenerator::onStopRequest()
{
    if(isRecording){
        requestStopRecording = true;
    }
}


bool MovieGenerator::doRecordingLoop()
{
    requestStopRecording = false;
    int frame = 0;
    double time = beginningTimeSpin.value();
    double endingTime = endingTimeCheck.isChecked() ? endingTimeSpin.value() : std::numeric_limits<double>::max();
    double timeStep = 1.0 / fpsSpin.value();
    TimeBar* timeBar = TimeBar::instance();
    SceneView* sceneView = SceneView::mainInstance();
    bool doContinue = true;

    filesystem::path directory(directoryEntry.string());
    filesystem::path basename(basenameEntry.string() + "%08u.png");

    if(directory.empty()){
        showWarningDialog(_("Please set a directory to output image files."));
        return false;

    } else {
        if(filesystem::exists(directory)){
            if(!filesystem::is_directory(directory)){
                showWarningDialog(fmt(_("%1% is not a directory.")) % directory);
                return false;
            }
        } else {
            filesystem::create_directories(directory);
        }
    }
    
    isRecording = true;

    format filenameFormat(fmt((directory / basename).string()));

    sceneView->setScreenSize(imageWidthSpin.value(), imageHeightSpin.value());

    while(time < endingTime && doContinue){

        doContinue = timeBar->setTime(time);

        MessageView::instance()->flush();

        if(requestStopRecording){
            break;
        }

        string filename = str(filenameFormat % frame);
        if(!sceneView->saveImage(filename)){
            showWarningDialog(fmt(_("Saving an image to \"%1%\" failed.")) % filename);
            requestStopRecording = true;
            break;
        }

        time += timeStep;
        frame++;
    }

    isRecording = false;

    return !requestStopRecording;
}
