/**
   \file
   \author Shin'ichiro Nakaoka
*/

#ifndef CNOID_BODY_SENSOR_H_INCLUDED
#define CNOID_BODY_SENSOR_H_INCLUDED

#include <cnoid/EigenTypes>
#include <cnoid/Referenced>
#include <string>
#include <vector>
#include <iostream>
#include <boost/function.hpp>
#include "exportdecl.h"

namespace cnoid {

    class Body;
    class Link;

    class CNOID_EXPORT Sensor
    {
      public:

        enum SensorType {
            COMMON = 0,
            FORCE,
            RATE_GYRO,
            ACCELERATION,
            PRESSURE,
            PHOTO_INTERRUPTER,
            VISION,
            TORQUE,
            RANGE,
            NUM_SENSOR_TYPES
        };

        static const int TYPE = COMMON;
		
        Sensor(); 
        virtual ~Sensor();

        static Sensor* create(int type);
        static void destroy(Sensor* sensor);

        void operator=(const Sensor& rhs);

        virtual void clear();
		
        std::string name;
        int type;
        int id;
        Link* link;
        Matrix3 localR;
        Vector3 localPos;

        virtual void putInformation(std::ostream& os);

    };


    class CNOID_EXPORT ForceSensor : public Sensor
    {
      public:
        static const int TYPE = FORCE;
		
        ForceSensor();

        void copyValuesFrom(const ForceSensor& source) {
            f = source.f;
            tau = source.tau;
        }
        
        Vector3 f;
        Vector3 tau;

        virtual void clear();
        virtual void putInformation(std::ostream& os);
    };


    class CNOID_EXPORT RateGyroSensor : public Sensor
    {
      public:
        static const int TYPE = RATE_GYRO;

        RateGyroSensor();

        void copyValuesFrom(const RateGyroSensor& source) {
            w = source.w;
        }
        
        Vector3 w;

        virtual void clear();
        virtual void putInformation(std::ostream& os);
    };


    class CNOID_EXPORT AccelSensor : public Sensor
    {
      public:
        EIGEN_MAKE_ALIGNED_OPERATOR_NEW
        
        static const int TYPE = ACCELERATION;

        AccelSensor();

        void copyValuesFrom(const AccelSensor& source) {
            dv = source.dv;
        }
        
        Vector3 dv;

        virtual void clear();
        virtual void putInformation(std::ostream& os);

        // The following members are used in the ForwardDynamics class
        Vector2 x[3]; 
    };

    class CNOID_EXPORT RangeSensor : public Sensor
    {
      public:
        static const int TYPE = RANGE;

        RangeSensor();

        double scanAngle, scanStep, scanRate, maxDistance;  
        std::vector<double> distances;
        double nextUpdateTime;
    };

    class SensorCalculatorImpl;

    //! \todo move the functions of this class into the Simulator class
    class CNOID_EXPORT SensorCalculator
    {
      public:
        SensorCalculator();
        ~SensorCalculator();

        typedef boost::function<void(ForceSensor* sensor, Vector6& out_wrench, Referenced* userData)> ForceFunction;
        
        void setForceFunction(ForceFunction func);
        void initialize(double timeStep, const Vector3& gravityAcceleration);
        void calcForceSensorValues(Body* body, Referenced* userData = 0);
        void calcGyroAndAccelSensorValues(Body* body);

      private:
        SensorCalculatorImpl* impl;
    };
        
    
};


#endif
