/*!
  @file
  @author Shin'ichiro Nakaoka
*/

#ifndef CNOID_BODYPLUGIN_SIMULATOR_ITEM_H_INCLUDED
#define CNOID_BODYPLUGIN_SIMULATOR_ITEM_H_INCLUDED

#include <cnoid/Item>
#include "exportdecl.h"

namespace cnoid {

    class Link;
    class Body;
    class BodyItem;
    class BodyMotionItem;
    class ControllerItem;
    class SimulationBodyImpl;
    class SimulatorItemImpl;

    class CNOID_EXPORT SimulationBody : public Referenced
    {
      public:
        SimulationBody();
        virtual ~SimulationBody();

        BodyItem* bodyItem() const;
        Body* body() const;
        ControllerItem* controller() const;

        /**
           Called from the simulation loop thread.
        */
        virtual void storeResult();

        virtual bool flushResult();
        
      private:
        SimulationBodyImpl* impl;
        friend class SimulatorItemImpl;
    };
    typedef boost::intrusive_ptr<SimulationBody> SimulationBodyPtr;
    
        
    class CNOID_EXPORT SimulatorItem : public Item
    {
      public:

        static void initialize(ExtensionManager* ext);

        enum RecordingMode { UNLIMITED = 0, SPECIFIC_TIME, TIMEBAR_RANGE, DIRECT, N_RECORDING_MODES };
        
        SimulatorItem();
        SimulatorItem(const SimulatorItem& org);
        virtual ~SimulatorItem();

        virtual double worldTimeStep();
            
        bool startSimulation(bool doReset = true);
        void stopSimulation();
        bool isRunning() const;
        int currentFrame() const;

        SignalProxy< boost::signal<void()> > sigSimulationFinished();

        RecordingMode recordingMode() const;

        virtual void setAllLinkPositionOutputMode(bool on);

        bool isAllLinkPositionOutputMode();

        virtual void selectMotionItems();
            
      protected:

        virtual void onDisconnectedFromRoot();

        virtual SimulationBody* createSimulationBody() = 0;

        virtual ControllerItem* createBodyMotionController(BodyItem* bodyItem, BodyMotionItem* bodyMotionItem);

        /**
           @param simBodies SimulatorBody objects which have a valid body
           @note This function is called from the main thread.
        */
        virtual bool initializeSimulation(std::vector<SimulationBody*>& simBodies) = 0;

        virtual void initializeSimulationThread();

        /**
           This function is called from the simulation loop thread.
           @param activeSimBodies SimulatorBody objects which are non-static ones.
        */
        virtual bool stepSimulation(std::vector<SimulationBody*>& activeSimBodies) = 0;

        /**
           \note This function is called from the main thread.
        */
        virtual void finalizeSimulation();

        void getColdetLinkPairs(std::vector<std::pair<Link*,Link*> >& out_pairs);

        void doPutProperties(PutPropertyFunction& putProperty);
        bool store(Archive& archive);
        bool restore(const Archive& archive);
            
      private:
            
        SimulatorItemImpl* impl;
        friend class SimulatorItemImpl;
    };
        
    typedef boost::intrusive_ptr<SimulatorItem> SimulatorItemPtr;
}

#endif
