/*!
  @file
  @author Shin'ichiro Nakaoka
*/

#include "ODESimulatorItem.h"
#include <cnoid/ItemManager>
#include <cnoid/Archive>
#include <cnoid/EigenUtil>
#include <cnoid/EigenArchive>
#include <cnoid/Body>
#include <cnoid/Link>
#include <cnoid/Sensor>
#include <cnoid/FloatingNumberString>
#include <boost/bind.hpp>
#include <boost/lexical_cast.hpp>
#include <ode/ode.h>
#include "gettext.h"

#include <iostream>

using namespace std;
using namespace boost;
using namespace cnoid;

namespace {

    const bool TRACE_FUNCTIONS = false;
    const double DEFAULT_GRAVITY_ACCELERATION = 9.80665;
    enum { STEP_ITERATIVE, STEP_BIG_MATRIX, NUM_STEP_MODES };

    typedef Eigen::Matrix<float, 3, 1> Vertex;

    struct Triangle {
        int indices[3];
    };

    dMatrix3 identity = {
        1.0, 0.0, 0.0, 0.0,
        0.0, 1.0, 0.0, 0.0,
        0.0, 0.0, 1.0, 0.0 };

    dMatrix3 flippedIdentity = {
        1.0,  0.0, 0.0, 0.0,
        0.0,  0.0, 1.0, 0.0,
        0.0, -1.0, 0.0, 0.0 };
    

    inline void makeInternal(Vector3& v) {
        double a = v.z();
        v.z() = -v.y();
        v.y() = a;
    }
    inline void toInternal(const Vector3& v, Vector3& out_v) {
        out_v << v.x(), v.z(), -v.y();
    }

    class ODEBody;

    class ODELink : public Referenced
    {
    public:
        Link* link;
        dBodyID bodyID;
        dJointID jointID;
        dGeomID geomID;
        dTriMeshDataID triMeshDataID;
        vector<Vertex> vertices;
        vector<Triangle> triangles;

        ODELink(ODESimulatorItemImpl* simImpl, ODEBody* odeBody, ODELink* parent,
                const Vector3& parentOrigin, Link* link);
        ~ODELink();
        void createLinkBody(ODESimulatorItemImpl* simImpl, dWorldID worldID, ODELink* parent, const Vector3& origin);
        void createGeometry(ODEBody* odeBody);
        void setKinematicStateToODE();
        void setKinematicStateToODEflip();
        void getKinematicStateFromODE();
        void getKinematicStateFromODEflip();
        void setTorqueToODE();
    };
    typedef boost::intrusive_ptr<ODELink> ODELinkPtr;


    class ODEBody : public SimulationBody
    {
    public:
        vector<ODELinkPtr> odeLinks;
        Body* body_;
        dWorldID worldID;
        dSpaceID spaceID;
        vector<dJointFeedback> forceSensorFeedbacks;

        ODEBody();
        ~ODEBody();
        void createBody(ODESimulatorItemImpl* simImpl);
        void setExtraJoints(bool flipYZ);
        void setKinematicStateToODE(bool flipYZ);
        void getKinematicStateFromODE(bool flipYZ);
        void setTorqueToODE();
        void alignToZAxisIn2Dmode();
    };
}


namespace cnoid {
  
    class ODESimulatorItemImpl
    {
    public:
        ODESimulatorItem* self;

        bool flipYZ;
        
        dWorldID worldID;
        dSpaceID spaceID;
        dJointGroupID contactJointGroupID;
        double timeStep;

        SensorCalculator sensorCalculator;
        
        Selection stepMode;
        Vector3 gravity;
        double friction;
        bool isJointLimitMode;
        bool is2Dmode;
        double globalERP;
        FloatingNumberString globalCFM;
        int numIterations;
        double overRelaxation;
        bool enableMaxCorrectingVel;
        FloatingNumberString maxCorrectingVel;
        double surfaceLayerDepth;

        ODESimulatorItemImpl(ODESimulatorItem* self);
        ODESimulatorItemImpl(ODESimulatorItem* self, const ODESimulatorItemImpl& org);
        void initialize();
        ~ODESimulatorItemImpl();
        void clear();
        bool initializeSimulation(std::vector<SimulationBody*>& simBodies);
        void addBody(ODEBody* odeBody);
        bool stepSimulation(std::vector<SimulationBody*>& activeSimBodies);
        static void getJointForce(ForceSensor* sensor, Vector6& out_wrench, Referenced* userData);
        static void getJointForceFlip(ForceSensor* sensor, Vector6& out_wrench, Referenced* userData);
        void doPutProperties(PutPropertyFunction& putProperty);
        void store(Archive& archive);
        void restore(const Archive& archive);
    };
}


ODELink::ODELink
(ODESimulatorItemImpl* simImpl, ODEBody* odeBody, ODELink* parent, const Vector3& parentOrigin, Link* link)
{
    odeBody->odeLinks.push_back(this);

    this->link = link;
    bodyID = 0;
    jointID = 0;
    geomID = 0;
    triMeshDataID = 0;
    
    Vector3 o = parentOrigin + link->b;
    
    if(odeBody->worldID){
        createLinkBody(simImpl, odeBody->worldID, parent, o);
    }
    if(link->coldetModel){
        createGeometry(odeBody);
    }

    for(Link* child = link->child; child; child = child->sibling){
        new ODELink(simImpl, odeBody, this, o, child);
    }
}


void ODELink::createLinkBody(ODESimulatorItemImpl* simImpl, dWorldID worldID, ODELink* parent, const Vector3& origin)
{
    bodyID = dBodyCreate(worldID);
    dBodySetData(bodyID, link);

    dMass mass;
    dMassSetZero(&mass);
    Matrix3& I = link->I;
    dMassSetParameters(&mass, link->m,
                       0.0, 0.0, 0.0,
                       I(0,0), I(1,1), I(2,2),
                       I(0,1), I(0,2), I(1,2));
    dBodySetMass(bodyID, &mass);

    Vector3 c;
    Vector3 o;
    Vector3 a;
    Vector3 d;

    if(!simImpl->flipYZ){
        c = link->c;
        o = origin;
        a = link->a;
        d = link->d;
        dBodySetRotation(bodyID, identity);
    } else {
        toInternal(link->c, c);
        toInternal(origin, o);
        toInternal(link->a, a);
        toInternal(link->d, d);
        dBodySetRotation(bodyID, flippedIdentity);
    }
        
    // set the default global position to set a joint
    Vector3 p = o + c;
    dBodySetPosition(bodyID, p.x(), p.y(), p.z());

    dBodyID parentBodyID = parent ? parent->bodyID : 0;

    switch(link->jointType){
        
    case Link::FIXED_JOINT:
        if(parentBodyID){
            jointID = dJointCreateFixed(worldID, 0);
            dJointAttach(jointID, bodyID, parentBodyID);
            dJointSetFixed(jointID);
        } else {
            dBodySetKinematic(bodyID);
        }
        break;
        
    case Link::FREE_JOINT:
        break;
        
    case Link::ROTATIONAL_JOINT:
        jointID = dJointCreateHinge(worldID, 0);
        dJointAttach(jointID, bodyID, parentBodyID);
        dJointSetHingeAnchor(jointID, o.x(), o.y(), o.z());
        dJointSetHingeAxis(jointID, a.x(), a.y(), a.z());
        if(simImpl->isJointLimitMode){
            if(link->ulimit < numeric_limits<double>::max()){
                dJointSetHingeParam(jointID, dParamHiStop, link->ulimit);
            }
            if(link->llimit > -numeric_limits<double>::max()){
                dJointSetHingeParam(jointID, dParamLoStop, link->llimit);
            }
        }
        break;
        
    case Link::SLIDE_JOINT:
        jointID = dJointCreateSlider(worldID, 0);
        dJointAttach(jointID, bodyID, parentBodyID);
        dJointSetSliderAxis(jointID, d.x(), d.y(), d.z());
        if(simImpl->isJointLimitMode){
            if(link->ulimit < numeric_limits<double>::max()){
                dJointSetSliderParam(jointID, dParamHiStop, link->ulimit);
            }
            if(link->llimit > -numeric_limits<double>::max()){
                dJointSetSliderParam(jointID, dParamLoStop, link->llimit);
            }
        }
        break;
    }
}


void ODELink::createGeometry(ODEBody* odeBody)
{
    ColdetModelPtr& cmodel = link->coldetModel;
    int numTriangles = cmodel->getNumTriangles();
    if(numTriangles <= 0){
        return;
    }

    int numVertices = cmodel->getNumVertices();
    vertices.resize(numVertices);
    const Vertex c(link->c.x(), link->c.y(), link->c.z());
    for(int i=0; i < numVertices; ++i){
        Vertex& v = vertices[i];
        cmodel->getVertex(i, v.x(), v.y(), v.z());
        v -= c;
    }
        
    triangles.resize(numTriangles);
    for(int i=0; i < numTriangles; ++i){
        Triangle& t = triangles[i];
        cmodel->getTriangle(i, t.indices[0], t.indices[1], t.indices[2]);
    }
    
    triMeshDataID = dGeomTriMeshDataCreate();
    dGeomTriMeshDataBuildSingle(triMeshDataID,
                                &vertices[0], sizeof(Vertex), vertices.size(),
                                &triangles[0], triangles.size() * 3, sizeof(Triangle));
    geomID = dCreateTriMesh(odeBody->spaceID, triMeshDataID, 0, 0, 0);
    dGeomSetBody(geomID, bodyID);
}


ODELink::~ODELink()
{
    if(jointID){
        dJointDestroy(jointID);
    }
    if(geomID){
        dGeomDestroy(geomID);
    }
    if(triMeshDataID){
        dGeomTriMeshDataDestroy(triMeshDataID);
    }
    if(bodyID){
        dBodyDestroy(bodyID);
    }
}


void ODELink::setKinematicStateToODE()
{
    const Matrix3& R = link->R;
    dMatrix3 R2 = { R(0,0), R(0,1), R(0,2), 0.0,
                    R(1,0), R(1,1), R(1,2), 0.0,
                    R(2,0), R(2,1), R(2,2), 0.0 };
    if(bodyID){
        dBodySetRotation(bodyID, R2);
        Vector3 lc = link->R * link->c;
        Vector3 c = link->p + lc;
        dBodySetPosition(bodyID, c.x(), c.y(), c.z());
        Vector3 v = link->v + link->w.cross(lc);
        dBodySetLinearVel(bodyID, v.x(), v.y(), v.z());
        dBodySetAngularVel(bodyID, link->w.x(), link->w.y(), link->w.z());

    } else if(geomID){
        Vector3 c = link->p + link->R * link->c;
        dGeomSetPosition(geomID, c.x(), c.y(), c.z());
        dGeomSetRotation(geomID, R2);
    }
}


void ODELink::setKinematicStateToODEflip()
{
    const Matrix3& R = link->R;
    dMatrix3 R2 = {  R(0,0),  R(0,1),  R(0,2), 0.0,
                     R(2,0),  R(2,1),  R(2,2), 0.0,
                    -R(1,0), -R(1,1), -R(1,2), 0.0 };
    if(bodyID){
        dBodySetRotation(bodyID, R2);
        Vector3 lc = link->R * link->c;
        Vector3 c = link->p + lc;
        dBodySetPosition(bodyID, c.x(), c.z(), -c.y());
        Vector3 v = link->v + link->w.cross(lc);
        dBodySetLinearVel(bodyID, v.x(), v.z(), -v.y());
        dBodySetAngularVel(bodyID, link->w.x(), link->w.z(), -link->w.y());

    } else if(geomID){
        Vector3 c = link->p + link->R * link->c;
        dGeomSetPosition(geomID, c.x(), c.z(), -c.y());
        dGeomSetRotation(geomID, R2);
    }
}


/**
   \note This method must not be called for a static body.
*/
void ODELink::getKinematicStateFromODE()
{
    if(jointID){
        if(link->jointType == Link::ROTATIONAL_JOINT){
            link->q = dJointGetHingeAngle(jointID);
            link->dq = dJointGetHingeAngleRate(jointID);
        } else if(link->jointType == Link::SLIDE_JOINT){
            link->q = dJointGetSliderPosition(jointID);
            link->dq = dJointGetSliderPositionRate(jointID);
        }
    }

    const dReal* R = dBodyGetRotation(bodyID);
    link->R <<
        R[0], R[1], R[2],
        R[4], R[5], R[6],
        R[8], R[9], R[10];

    typedef Eigen::Map<const Eigen::Matrix<dReal, 3, 1> > toVector3;
    Vector3 c = link->R * link->c;
    link->p = toVector3(dBodyGetPosition(bodyID)) - c;
    link->w = toVector3(dBodyGetAngularVel(bodyID));
    link->v = toVector3(dBodyGetLinearVel(bodyID)) - link->w.cross(c);
}


/**
   \note This method must not be called for a static body.
*/
void ODELink::getKinematicStateFromODEflip()
{
    if(jointID){
        if(link->jointType == Link::ROTATIONAL_JOINT){
            link->q = dJointGetHingeAngle(jointID);
            link->dq = dJointGetHingeAngleRate(jointID);
        } else if(link->jointType == Link::SLIDE_JOINT){
            link->q = dJointGetSliderPosition(jointID);
            link->dq = dJointGetSliderPositionRate(jointID);
        }
    }

    const dReal* R = dBodyGetRotation(bodyID);
    link->R <<
         R[0],  R[1],  R[2],
        -R[8], -R[9], -R[10],
         R[4],  R[5],  R[6];
    Vector3 c;
    toInternal(link->R * link->c, c);
    
    typedef Eigen::Map<const Eigen::Matrix<dReal, 3, 1> > toVector3;
    Vector3 p = toVector3(dBodyGetPosition(bodyID)) - c;
    toVector3 w(dBodyGetAngularVel(bodyID));
    Vector3 v = toVector3(dBodyGetLinearVel(bodyID)) - w.cross(c);
    
    link->p << p.x(), -p.z(), p.y();
    link->w << w.x(), -w.z(), w.y(); 
    link->v << v.x(), -v.z(), v.y();
}


/**
   \note This method must not be called for the root link and a static body.
*/
void ODELink::setTorqueToODE()
{
    if(link->jointType == Link::ROTATIONAL_JOINT){
        dJointAddHingeTorque(jointID, link->u);
    } else if(link->jointType == Link::SLIDE_JOINT){
        dJointAddSliderForce(jointID, link->u);
    }
}


ODEBody::ODEBody()
{
    body_ = 0;
    worldID = 0;
    spaceID = 0;
}


ODEBody::~ODEBody()
{
    if(spaceID){
        dSpaceDestroy(spaceID);
    }
}


void ODEBody::createBody(ODESimulatorItemImpl* simImpl)
{
    body_ = body();
    worldID = body_->isStaticModel() ? 0 : simImpl->worldID;
    spaceID = dHashSpaceCreate(simImpl->spaceID);
    dSpaceSetCleanup(spaceID, 0);

    ODELink* rootLink = new ODELink(simImpl, this, 0, Vector3::Zero(), body_->rootLink());

    // set joint feedbacks for force sensors
    int n = body_->numSensors(Sensor::FORCE);
    forceSensorFeedbacks.resize(n);
    for(int i=0; i < n; ++i){
        ForceSensor* sensor = body_->sensor<ForceSensor>(i);
        Link* link = sensor->link;
        dJointSetFeedback(odeLinks[link->index]->jointID, &forceSensorFeedbacks[i]);
    }

    setKinematicStateToODE(simImpl->flipYZ);

    setExtraJoints(simImpl->flipYZ);

    if(simImpl->is2Dmode && worldID){
        dJointID planeJointID = dJointCreatePlane2D(worldID, 0);
        dJointAttach(planeJointID, rootLink->bodyID, 0);
    }
    
    setTorqueToODE();
}


void ODEBody::setExtraJoints(bool flipYZ)
{
    int n = body_->extraJoints.size();

    for(int j=0; j < n; ++j){

        Body::ExtraJoint& extraJoint = body_->extraJoints[j];

        ODELinkPtr odeLinkPair[2];
        for(int i=0; i < 2; ++i){
            ODELinkPtr odeLink;
            Link* link = extraJoint.link[i];
            if(link->index < odeLinks.size()){
                odeLink = odeLinks[link->index];
                if(odeLink->link == link){
                    odeLinkPair[i] = odeLink;
                }
            }
            if(!odeLink){
                break;
            }
        }

        if(odeLinkPair[1]){

            dJointID jointID = 0;
            Link* link = odeLinkPair[0]->link;
            Vector3 p = link->segmentAttitude() * extraJoint.point[0] + link->p;
            Vector3 a = link->segmentAttitude() * extraJoint.axis;
            if(flipYZ){
                makeInternal(p);
                makeInternal(a);
            }

            // \todo do the destroy management for these joints
            if(extraJoint.type == Body::EJ_PISTON){
                jointID = dJointCreatePiston(worldID, 0);
                dJointAttach(jointID, odeLinkPair[0]->bodyID, odeLinkPair[1]->bodyID);
                dJointSetPistonAnchor(jointID, p.x(), p.y(), p.z());
                dJointSetPistonAxis(jointID, a.x(), a.y(), a.z());

            } else if(extraJoint.type == Body::EJ_BALL){
                jointID = dJointCreateBall(worldID, 0);
                dJointAttach(jointID, odeLinkPair[0]->bodyID, odeLinkPair[1]->bodyID);
                dJointSetBallAnchor(jointID, p.x(), p.y(), p.z());
            }
        }
    }
}


void ODEBody::setKinematicStateToODE(bool flipYZ)
{
    if(!flipYZ){
        for(size_t i=0; i < odeLinks.size(); ++i){
            odeLinks[i]->setKinematicStateToODE();
        }
    } else {
        for(size_t i=0; i < odeLinks.size(); ++i){
            odeLinks[i]->setKinematicStateToODEflip();
        }
    }
}


void ODEBody::getKinematicStateFromODE(bool flipYZ)
{
    if(!flipYZ){
        for(size_t i=0; i < odeLinks.size(); ++i){
            odeLinks[i]->getKinematicStateFromODE();
        }
    } else {
        for(size_t i=0; i < odeLinks.size(); ++i){
            odeLinks[i]->getKinematicStateFromODEflip();
        }
    }
}


void ODEBody::setTorqueToODE()
{
    // Skip the root link
    for(size_t i=1; i < odeLinks.size(); ++i){
        odeLinks[i]->setTorqueToODE();
    }
}


void ODEBody::alignToZAxisIn2Dmode()
{
    static const Quat r(AngleAxis(PI / 2.0, Vector3(1.0, 0.0, 0.0)));

    dBodyID& bodyID = odeLinks.front()->bodyID;

    const dReal* q0 = dBodyGetQuaternion(bodyID);
    Quat q(q0[0], q0[1], q0[2], q0[3]);
    Quat q2 = r * q;
    q2.x() = 0.0;
    q2.z() = 0.0;
    q2.normalize();
    Quat q3 = r.inverse() * q2;
    dReal q4[4];
    q4[0] = q3.w();
    q4[1] = q3.x();    
    q4[2] = q3.y();    
    q4[3] = q3.z();    
    dBodySetQuaternion(bodyID, q4);

    const dReal* w = dBodyGetAngularVel(bodyID);
    dBodySetAngularVel(bodyID, 0, 0, w[2]);
}


void ODESimulatorItem::initialize(ExtensionManager* ext)
{
    ext->itemManager().registerClass<ODESimulatorItem>(N_("ODESimulatorItem"));
    ext->itemManager().addCreationPanel<ODESimulatorItem>();
}


ODESimulatorItem::ODESimulatorItem()
{
    impl = new ODESimulatorItemImpl(this);
}


ODESimulatorItemImpl::ODESimulatorItemImpl(ODESimulatorItem* self)
    : self(self),
      stepMode(NUM_STEP_MODES, CNOID_GETTEXT_DOMAIN_NAME)

{
    initialize();

    stepMode.setSymbol(STEP_ITERATIVE,  N_("Iterative (quick step)"));
    stepMode.setSymbol(STEP_BIG_MATRIX, N_("Big matrix"));
    stepMode.select(STEP_ITERATIVE);
    
    gravity << 0.0, 0.0, -DEFAULT_GRAVITY_ACCELERATION;
    globalERP = 0.4;
    globalCFM = "1.0e-10";
    numIterations = 50;
    overRelaxation = 1.3;
    enableMaxCorrectingVel = true;
    maxCorrectingVel = "1.0e-3";
    surfaceLayerDepth = 0.0001;
    friction = 1.0;
    isJointLimitMode = false;
    is2Dmode = false;
    flipYZ = false;

    
}


ODESimulatorItem::ODESimulatorItem(const ODESimulatorItem& org)
    : SimulatorItem(org)
{
    impl = new ODESimulatorItemImpl(this, *org.impl);
}


ODESimulatorItemImpl::ODESimulatorItemImpl(ODESimulatorItem* self, const ODESimulatorItemImpl& org)
    : self(self)
{
    initialize();

    stepMode = org.stepMode;
    gravity = org.gravity;
    globalERP = org.globalERP;
    globalCFM = org.globalCFM;
    numIterations = org.numIterations;
    overRelaxation = org.overRelaxation;
    enableMaxCorrectingVel = org.enableMaxCorrectingVel;
    maxCorrectingVel = org.maxCorrectingVel;
    surfaceLayerDepth = org.surfaceLayerDepth;
    friction = org.friction;
    isJointLimitMode = org.isJointLimitMode;
    is2Dmode = org.is2Dmode;
    flipYZ = org.flipYZ;
}


void ODESimulatorItemImpl::initialize()
{
    worldID = 0;
    spaceID = 0;
    contactJointGroupID = dJointGroupCreate(0);
    self->SimulatorItem::setAllLinkPositionOutputMode(true);
}


ODESimulatorItem::~ODESimulatorItem()
{
    delete impl;
}


ODESimulatorItemImpl::~ODESimulatorItemImpl()
{
    clear();

    if(contactJointGroupID){
        dJointGroupDestroy(contactJointGroupID);
    }
}


void ODESimulatorItem::setAllLinkPositionOutputMode(bool on)
{
    // The mode is not changed.
    // This simulator only supports the all link position output
    // because joint positions may be slightly changed
}


void ODESimulatorItemImpl::clear()
{
    dJointGroupEmpty(contactJointGroupID);

    if(worldID){
        dWorldDestroy(worldID);
        worldID = 0;
    }
    if(spaceID){
        dSpaceDestroy(spaceID);
        spaceID = 0;
    }
}    


ItemPtr ODESimulatorItem::doDuplicate() const
{
    return new ODESimulatorItem(*this);
}


SimulationBody* ODESimulatorItem::createSimulationBody()
{
    return new ODEBody();
}


bool ODESimulatorItem::initializeSimulation(std::vector<SimulationBody*>& simBodies)
{
    return impl->initializeSimulation(simBodies);
}


bool ODESimulatorItemImpl::initializeSimulation(std::vector<SimulationBody*>& simBodies)
{
    clear();

    flipYZ = is2Dmode;

    Vector3 g = gravity;
    
    if(!flipYZ){
        sensorCalculator.setForceFunction(bind(&ODESimulatorItemImpl::getJointForce, _1, _2, _3));
    } else {
        toInternal(gravity, g);
        sensorCalculator.setForceFunction(bind(&ODESimulatorItemImpl::getJointForceFlip, _1, _2, _3));
    }

    worldID = dWorldCreate();
    spaceID = dHashSpaceCreate(0);
    dSpaceSetCleanup(spaceID, 0);

    dRandSetSeed(0);
    dWorldSetGravity(worldID, g.x(), g.y(), g.z());
    dWorldSetERP(worldID, globalERP);
    dWorldSetCFM(worldID, globalCFM.value());
    dWorldSetContactSurfaceLayer(worldID, 0.0);
    dWorldSetQuickStepNumIterations(worldID, numIterations);
    dWorldSetQuickStepW(worldID, overRelaxation);
    dWorldSetContactMaxCorrectingVel(worldID, enableMaxCorrectingVel ? maxCorrectingVel.value() : dInfinity);
    dWorldSetContactSurfaceLayer(worldID, surfaceLayerDepth);

    timeStep = self->worldTimeStep();

    for(size_t i=0; i < simBodies.size(); ++i){
        addBody(static_cast<ODEBody*>(simBodies[i]));
    }

    sensorCalculator.initialize(timeStep, g);

    return true;
}


void ODESimulatorItemImpl::addBody(ODEBody* odeBody)
{
    Body* body = odeBody->body();

    Link* rootLink = body->rootLink();
    rootLink->v.setZero();
    rootLink->dv.setZero();
    rootLink->w.setZero();
    rootLink->dw.setZero();

    for(int i=0; i < body->numJoints(); ++i){
        Link* joint = body->joint(i);
        joint->u = 0.0;
        joint->dq = 0.0;
        joint->ddq = 0.0;
    }
    
    body->clearExternalForces();
    body->calcForwardKinematics(true, true);

    odeBody->createBody(this);
}


void ODESimulatorItem::initializeSimulationThread()
{
    dAllocateODEDataForThread(dAllocateMaskAll);
}


static void nearCallback(void* data, dGeomID g1, dGeomID g2)
{
    if(dGeomIsSpace(g1) || dGeomIsSpace(g2)) { 
        dSpaceCollide2(g1, g2, data, &nearCallback);
        if(false) { // Currently just skip same body link pairs. 
            if(dGeomIsSpace(g1)){
                dSpaceCollide((dSpaceID)g1, data, &nearCallback);
            }
            if(dGeomIsSpace(g2)){
                dSpaceCollide((dSpaceID)g2, data, &nearCallback);
            }
        }
    } else {
        ODESimulatorItemImpl* impl = (ODESimulatorItemImpl*)data;
        static const int MaxNumContacts = 100;
        dContact contacts[MaxNumContacts];
        int numContacts = dCollide(g1, g2, MaxNumContacts, &contacts[0].geom, sizeof(dContact));

        if(numContacts > 0){
            dBodyID body1ID = dGeomGetBody(g1);
            dBodyID body2ID = dGeomGetBody(g2);
            for(int i=0; i < numContacts; ++i){
                dSurfaceParameters& surface = contacts[i].surface;
                //surface.mode = dContactApprox1 | dContactBounce;
                //surface.bounce = 0.0;
                //surface.bounce_vel = 1.0;
                surface.mode = dContactApprox1;
                surface.mu = impl->friction;
                dJointID jointID = dJointCreateContact(impl->worldID, impl->contactJointGroupID, &contacts[i]);
                dJointAttach(jointID, body1ID, body2ID);
            }
        }
    }
}


bool ODESimulatorItem::stepSimulation(std::vector<SimulationBody*>& activeSimBodies)
{
    return impl->stepSimulation(activeSimBodies);
}


bool ODESimulatorItemImpl::stepSimulation(std::vector<SimulationBody*>& activeSimBodies)
{
    for(size_t i=0; i < activeSimBodies.size(); ++i){
        ODEBody* odeBody = static_cast<ODEBody*>(activeSimBodies[i]);
        odeBody->body_->setVirtualJointForces();
        odeBody->setTorqueToODE();
    }

    dJointGroupEmpty(contactJointGroupID);
    dSpaceCollide(spaceID, (void*)this, &nearCallback);

    if(stepMode.is(STEP_ITERATIVE)){
        dWorldQuickStep(worldID, timeStep);
    } else {
        dWorldStep(worldID, timeStep);
    }

    //! \todo Bodies with sensors should be managed by the specialized container to increase the efficiency
    for(size_t i=0; i < activeSimBodies.size(); ++i){
        ODEBody* odeBody = static_cast<ODEBody*>(activeSimBodies[i]);
        if(is2Dmode){
            odeBody->alignToZAxisIn2Dmode();
        }
        sensorCalculator.calcForceSensorValues(odeBody->body_, odeBody);
        odeBody->getKinematicStateFromODE(flipYZ);
        sensorCalculator.calcGyroAndAccelSensorValues(odeBody->body_);
    }

    return true;
}


void ODESimulatorItemImpl::getJointForce(ForceSensor* sensor, Vector6& out_wrench, Referenced* userData)
{
    ODEBody* odeBody = static_cast<ODEBody*>(userData);
    dJointFeedback& fb = odeBody->forceSensorFeedbacks[sensor->id];

    out_wrench <<
        fb.f2[0], fb.f2[1], fb.f2[2],
        fb.t2[0], fb.t2[1], fb.t2[2];
}


void ODESimulatorItemImpl::getJointForceFlip(ForceSensor* sensor, Vector6& out_wrench, Referenced* userData)
{
    ODEBody* odeBody = static_cast<ODEBody*>(userData);
    dJointFeedback& fb = odeBody->forceSensorFeedbacks[sensor->id];

    out_wrench <<
        fb.f2[0], -fb.f2[2], fb.f2[1],
        fb.t2[0], -fb.t2[2], fb.t2[1];
}


void ODESimulatorItem::doPutProperties(PutPropertyFunction& putProperty)
{
    SimulatorItem::doPutProperties(putProperty);
    impl->doPutProperties(putProperty);
}


void ODESimulatorItemImpl::doPutProperties(PutPropertyFunction& putProperty)
{
    putProperty(_("Step mode"), stepMode, changeProperty(stepMode));

    putProperty(_("Gravity"), str(gravity), bind(toVector3, _1, ref(gravity)));

    putProperty.decimals(2).min(0.0)
        (_("Friction"), friction, changeProperty(friction));

    putProperty(_("Limit joint range"), isJointLimitMode, changeProperty(isJointLimitMode));

    putProperty.decimals(1).min(0.0).max(1.0)
        (_("Global ERP"), globalERP, changeProperty(globalERP));

    putProperty(_("Global CFM"), globalCFM,
                bind(&FloatingNumberString::setNonNegativeValue, ref(globalCFM), _1));

    putProperty.min(1)
        (_("Iterations"), numIterations, changeProperty(numIterations));

    putProperty.min(0.1).max(1.9)
        (_("Over relaxation"), overRelaxation, changeProperty(overRelaxation));

    putProperty(_("Limit correcting vel."), enableMaxCorrectingVel, changeProperty(enableMaxCorrectingVel));

    putProperty(_("Max correcting vel."), maxCorrectingVel,
                bind(&FloatingNumberString::setNonNegativeValue, ref(maxCorrectingVel), _1));

    putProperty(_("2D mode"), is2Dmode, changeProperty(is2Dmode));
}


bool ODESimulatorItem::store(Archive& archive)
{
    SimulatorItem::store(archive);
    impl->store(archive);
    return true;
}


void ODESimulatorItemImpl::store(Archive& archive)
{
    archive.write("stepMode", stepMode.selectedSymbol());
    writeMatrix(archive, "gravity", gravity);
    archive.write("friction", friction);
    archive.write("jointLimitMode", isJointLimitMode);
    archive.write("globalERP", globalERP);
    archive.write("globalCFM", globalCFM);
    archive.write("numIterations", numIterations);
    archive.write("overRelaxation", overRelaxation);
    archive.write("limitCorrectingVel", enableMaxCorrectingVel);
    archive.write("maxCorrectingVel", maxCorrectingVel);
    archive.write("2Dmode", is2Dmode);
}


bool ODESimulatorItem::restore(const Archive& archive)
{
    SimulatorItem::restore(archive);
    impl->restore(archive);
    return true;
}


void ODESimulatorItemImpl::restore(const Archive& archive)
{
    string symbol;
    if(archive.read("stepMode", symbol)){
        stepMode.select(symbol);
    }
    read(archive, "gravity", gravity);
    archive.read("friction", friction);
    archive.read("jointLimitMode", isJointLimitMode);
    archive.read("globalERP", globalERP);
    globalCFM = archive.get("globalCFM", globalCFM.string());
    archive.read("numIterations", numIterations);
    archive.read("overRelaxation", overRelaxation);
    archive.read("limitCorrectingVel", enableMaxCorrectingVel);
    maxCorrectingVel = archive.get("maxCorrectingVel", maxCorrectingVel.string());
    archive.read("2Dmode", is2Dmode);
}
